from skimage.measure import label, regionprops, regionprops_table
from skimage.feature import canny
from skimage import data, color
from skimage.transform import hough_circle, hough_circle_peaks
from skimage.draw import circle_perimeter
from skimage.util import img_as_ubyte
from skimage import filters
from skimage import exposure
from skimage import feature
from skimage.color import rgb2gray
from skimage.draw import disk, polygon
import skimage as ski
import scipy.stats as stat
import numpy as np
import os, sys
import time
import glob


def find_edges( image_to_process , rescale_percentile_low , rescale_percentile_hi , sigma ):
    """
    1) take raw image, rescale at low and high bounds
    2) return edges
    """
    s_low , s_hi = np.percentile( image_to_process , ( rescale_percentile_low , rescale_percentile_hi ) )
    rescaled = exposure.rescale_intensity( image_to_process , in_range=( s_low , s_hi ) )
    return canny( rescaled , sigma=sigma )


def rescale( image_to_process , rescale_percentile_low , rescale_percentile_hi ):
    """
    1) take raw image, rescale at low and high bounds
    """
    s_low , s_hi = np.percentile( image_to_process , ( rescale_percentile_low , rescale_percentile_hi ) ) 
    return exposure.rescale_intensity( image_to_process , in_range=( s_low , s_hi ) )


def platform_mask( edges , radius_low , radius_hi ):
    """
    1) take a source Morris Water Maze image and find a circle of radius in the low-to-hi interval px
    --> to estimate the interval plot an image of your MWM pool and see approximately how large in px the MWM platform radius is
    --> in our case the interval was low=15, hi=40
    2) return a 2D mask array
    """
    # Detect two radii
    hough_radii = np.arange( radius_low , radius_hi , 1 )
    hough_res = hough_circle( edges , hough_radii )
    # Select the most prominent 3 circles
    accums , cx , cy , radii = hough_circle_peaks( hough_res , hough_radii , total_num_peaks=1 )
    cx = sorted( cx )
    cy = sorted( cy )
    radii = sorted( radii )
    mask = np.zeros( ( len(edges) , len(edges[0]) , 3 ) )
    rr , cc = disk( ( cy[ 0 ] , cx[ 0 ] ) , radii[ 0 ] , shape=mask.shape )
    mask[ rr , cc ] = ( 1 , 1 , 0 )
    return rgb2gray(mask) , radii[0]


def pool_mask( edges , radius_low , radius_hi ):
    """
    1) take a source Morris Water Maze image and find a circle of radius in the low-to-hi interval px
    --> to estimate the interval plot an image of your MWM pool and see approximately how large in px the MWM platform radius is
    --> in our case the interval was low=300, hi=550
    2) return a 2D mask array
    """
    # Detect two radii
    hough_radii = np.arange( radius_low , radius_hi , 50 )
    hough_res = hough_circle( edges , hough_radii )
    # Select the most prominent 3 circles
    accums , cx , cy , radii = hough_circle_peaks( hough_res , hough_radii , total_num_peaks=1 )
    cx = sorted( cx )
    cy = sorted( cy )
    radii = sorted( radii )
    mask = np.zeros( ( len(edges) , len(edges[0]) , 3 ) )
    rr , cc = disk( ( cy[ 0 ] , cx[ 0 ] ) , radii[ 0 ] , shape=mask.shape )
    mask[ rr , cc ] = ( 1 , 1 , 0 )
    return rgb2gray(mask) , radii[0] , cx[0] , cy[0]


def find_coordinates( img ):
        """
        take source image, find coords of the prominent edges object
        """
        label_img = label( img )
        regions = regionprops( label_img )
        for props in regions:
            centroid_y , centroid_x = props.centroid
        return centroid_x , centroid_y


def movement_stats( list_of_x_coordinates , list_of_y_coordinates , platform_diameter_cm , platform_radius , fps ):
    """
    1) define the diameter of the platform in cm; give x, y coordinates of the mouse's trace
    3) take the platform radius from the platform mask global variable and calculate the px to cm conversion
    2) return the distance and duration of the swim
    """
    platform_diameter_px = platform_radius * 2 #because d=2r
    distance_index = platform_diameter_cm / platform_diameter_px
    i = 0
    distance = 0
    while i < len(list_of_x_coordinates)-1:
        a = ( list_of_x_coordinates[ i ] - list_of_x_coordinates[ i + 1 ] )
        b = ( list_of_y_coordinates[ i ] - list_of_y_coordinates[ i + 1 ] )
        distance += np.sqrt( a**2 + b**2 )
        i += 1
    distance_cm = np.round( distance * distance_index , 0 ).astype(int)
    print( distance_cm , "cm swum in" , len(list_of_x_coordinates )/fps , "s")
    return distance_cm , len(list_of_x_coordinates )/fps


def trim_start( list_of_video_frames , pool_mask , rescale_percentile_low , rescale_percentile_hi ):
    """
    1) take list of water maze frames, rescale intensity (in our case set to low=60, hi=70)
    2) take sums of the pixel values in each frame (masked for pool)
    3) find global minimum - assuming that's the maximum overlap between hands in gloves and the pool
    4) find the 1st value above the median of the list between the minimum and the list end - assuming that's where the hands are fully retracted
    5) return trimmed list 
    """
    print("[Video trimmer]: WARNING - for short swims manual selection of frames is advised")

    vals = []
    for img in list_of_video_frames:
        current_frame = img * pool_mask
        current_rescaled = rescale( current_frame , 60 , 70 )
        vals.append( np.sum(current_rescaled) )
    print("[Video trimmer]: rescaling frames ... \t\t done")

    if len(vals) <= 700:
        counts, bins = np.histogram(vals,bins=10)
    elif len(vals) > 400:
        counts, bins = np.histogram(vals[ : np.round( len(vals)*0.5 , 0 ).astype(int) ],bins=10)
    else: pass    
    start = ""
    start_index = 0
    while ( start != "found" ) and ( start_index < len(vals)-1 ):
        lower_bound = np.round( bins[ np.argmax(counts)-2 ] , 0 )
        upper_bound = np.round( np.mean( [ bins[ np.argmax(counts)+2 ] , bins[ np.argmax(counts)+3 ] ] ) , 0 )
        reasonable_range = np.arange( lower_bound , upper_bound )
        current_val = np.round( vals[ start_index ] , 0 )
        consecutive_vals = np.round( vals[ start_index : start_index+30 ] , 0 )
        bool_list = []
        for val in consecutive_vals:
            if val in reasonable_range:
                bool_list.append(True)
            else: bool_list.append(False)
        if all(bool_list) == True:
            start = "found"
            trimmed_list = list_of_video_frames[ start_index : ]
        else: start_index += 1
    print("[Video trimmer]: \n\t\t trim start sum of pixels:\t" , current_val ,
          "\n\t\t median pixel sum:\t\t" , np.round( np.median(vals) , 0 ) ,
          "\n\t\t video trimmed at frame:\t" , start_index , "; is" , len(trimmed_list) , "frames long")
    return trimmed_list


from matplotlib.collections import LineCollection
from matplotlib.colors import ListedColormap
from matplotlib import pyplot as plt
import matplotlib.lines as lines
import numpy as np
import os, sys
import time
import glob
import cv2

def video_to_frames( vid_path , output_path , output_fps ):
    """Function to extract frames from input video file
    and save them as separate frames in an output directory.
    Args:
        vid_path: src video file path
        output_fps: desired frame rate
    Returns:
        None
    """
    extention = "."+vid_path[-3:]
    folder = os.path.basename( vid_path ).replace( extention , "" )
    output_loc = output_path + folder
    if folder not in glob.glob( output_path ):
        try:
            os.mkdir( output_loc )
            print("[Video to frames: creating output folder ... \t done" )
        except OSError:
            pass
    # Log the time
    time_start = time.time()
    # Start capturing the feed
    cap = cv2.VideoCapture( vid_path )
    # Find the number of frames
    video_length = int( cap.get( cv2.CAP_PROP_FRAME_COUNT ) ) - 1
    cap.set( cv2.CAP_PROP_FPS , output_fps )
    print("[Video to frames]: FPS reset to: \t\t" , output_fps )
    count = 0
    # Start converting the video
    while cap.isOpened():
        # Extract the frame
        ret, frame = cap.read()
        if not ret:
            continue
        # Write the results back to output location.
        cv2.imwrite( output_loc + "/%#05d.jpg" % (count+output_fps), frame )
        count += 1
        cap.set( cv2.CAP_PROP_POS_FRAMES , count )
        # If there are no more frames left
        if ( count > ( video_length-1) ):
            # Log the time again
            time_end = time.time()
            # Release the feed
            cap.release()
            break
    print("[Video to frames]: done" )

def colored_line(x, y, c, ax, **lc_kwargs):
    """
    Plot a line with a color specified along the line by a third value.

    It does this by creating a collection of line segments. Each line segment is
    made up of two straight lines each connecting the current (x, y) point to the
    midpoints of the lines connecting the current point with its two neighbors.
    This creates a smooth line with no gaps between the line segments.

    Parameters
    ----------
    x, y : array-like
        The horizontal and vertical coordinates of the data points.
    c : array-like
        The color values, which should be the same size as x and y.
    ax : Axes
        Axis object on which to plot the colored line.
    **lc_kwargs
        Any additional arguments to pass to matplotlib.collections.LineCollection
        constructor. This should not include the array keyword argument because
        that is set to the color argument. If provided, it will be overridden.

    Returns
    -------
    matplotlib.collections.LineCollection
        The generated line collection representing the colored line.
    """
    if "array" in lc_kwargs:
        warnings.warn('The provided "array" keyword argument will be overridden')

    # Default the capstyle to butt so that the line segments smoothly line up
    default_kwargs = {"capstyle": "butt"}
    default_kwargs.update(lc_kwargs)

    # Compute the midpoints of the line segments. Include the first and last points
    # twice so we don't need any special syntax later to handle them.
    x = np.asarray(x)
    y = np.asarray(y)
    x_midpts = np.hstack((x[0], 0.5 * (x[1:] + x[:-1]), x[-1]))
    y_midpts = np.hstack((y[0], 0.5 * (y[1:] + y[:-1]), y[-1]))

    # Determine the start, middle, and end coordinate pair of each line segment.
    # Use the reshape to add an extra dimension so each pair of points is in its
    # own list. Then concatenate them to create:
    # [
    #   [(x1_start, y1_start), (x1_mid, y1_mid), (x1_end, y1_end)],
    #   [(x2_start, y2_start), (x2_mid, y2_mid), (x2_end, y2_end)],
    #   ...
    # ]
    coord_start = np.column_stack((x_midpts[:-1], y_midpts[:-1]))[:, np.newaxis, :]
    coord_mid = np.column_stack((x, y))[:, np.newaxis, :]
    coord_end = np.column_stack((x_midpts[1:], y_midpts[1:]))[:, np.newaxis, :]
    segments = np.concatenate((coord_start, coord_mid, coord_end), axis=1)

    lc = LineCollection(segments, **default_kwargs)
    lc.set_array(c)  # set the colors of each segment

    return ax.add_collection(lc)


from matplotlib.colors import ListedColormap
from matplotlib import pyplot as plt
import matplotlib.lines as lines
from skimage import exposure
from skimage import feature
import skimage as ski
import scipy.stats as stat
import numpy as np
import os, sys
import time
import glob


def analyse_sequnce( positions ):
    """
    Takes a Spontaneous alteration sequence and returns the SA index
    """
    alternation_count = 0
    i = 0
    while i < len(positions)-2:
        #check if 3 consecutive positions are all different, i++ if so
        if ( positions[ i ] != positions[ i + 1 ] ) and ( positions[ i ] != positions[ i + 2 ] ) and ( positions[ i + 1 ] != positions[ i + 2 ] ):
            alternation_count += 1
        else: pass
        i += 1
    #do the maths, prepared for div by 0
    try: index = alternation_count / ( len(positions) - 2 )
    except ZeroDivisionError:
        index = 0
    return index , alternation_count


def ymaze_positions( subfolder , fps , rescale_low , rescale_hi , sigma ):
    """
    Spontaneous alternation analysis tool. Returns subfolder path (ie subject name), positions detected, and respective timestamps.
    1) loads .jpg frames from a subfolder, assumes .png masks for Y maze arms are in "/mask" subfolder
    2) makes sure all images are 2D arrays and masks are binary, rescales intensity (low=0, hi=5, sigma=5)
    3) each frame is multiplied by masks and results are compared - largest array sum assumed to be mouse's position
    4) position and respective timestamp (assuming framerate set to fps arg) are saved to list
    """
    #load images and masks
    image_list = []
    mask_list = []
    for frame in glob.glob(subfolder+"*.jpg"):
        im = np.sum( ski.io.imread( frame ) , axis=2 )
        image_list.append( im )
    for file in glob.glob( subfolder + "/masks/*.png" ):
        #load masks and threshold them
        mask_list.append( np.where( np.sum( ski.io.imread( file ) , axis=2) < 1000 , 0 , 1 ) )
    #IMAGE PROCESSING
    sequence = []
    overlaps = []
    timestamps = []
    positions = ["A","B","C","D"]
    position = "x"
    timestamp = 0
    #load one frame from image list
    for frame in image_list:
        #increment timestamp by fps of the video
        timestamp += fps
        #rescale intensity, in our case low=0, hi=5
        s_low, s_hi = np.percentile( frame , ( rescale_low , rescale_hi ) )
        rescaled = exposure.rescale_intensity( frame, in_range=( s_low, s_hi ) )
        #find edges in current rescaled img, in our case sigma=7
        edges = feature.canny( rescaled, sigma=sigma )
        #overlap (1*1) between image and arms' masks, outside mask = 0 ie 1*0 or 0*0
        indices = []        
        for mask in mask_list:
            indices.append( np.sum( edges * mask ) )
        #find greatest overlap, ie highest area ratio
        index_max = sorted( indices )[ 3 ]
        i = 0
        for index in indices:
            if all( [ index == index_max , ( sum(indices)-index_max ) == 0 ] ) == True:
                position = positions[ i ]
            elif indices[3] == index_max:
                position = "D"
            else: pass
            i += 1
        #save position if previous frame different
        if ( len(sequence) == 0 ) or position != sequence[ len(sequence)-1 ]:
            sequence.append( position )
            timestamps.append( timestamp )
        else: pass
    try: sequence.remove("x")
    except: pass
    #remove all Ds and respective timestamps
    respective_tstamp = 0
    for p in sequence:
        if p == "D":
            sequence.remove("D")
            timestamps.remove( timestamps[ respective_tstamp ] )
        else:
            pass
        respective_tstamp += 1
    return subfolder , sequence , timestamps